from __future__ import absolute_import
import re
import yaml
import os
from typing import Dict

from log4py import logging
log = logging.getLogger('python.lib.configure.yamlLoader')

pattern = re.compile(r"\${([^{}]+)}")


def get_yaml_objects(file: str, params: Dict[str, str]):
    if not os.path.exists(file):
        return []

    try:
        yaml_text = _prapare_variable(file, params)
        yaml_iter = yaml.safe_load_all(yaml_text)
        return list(yaml_iter)
    except Exception:
        log.error("Error loading YAML configuration file: %s" % file)
        raise


def _prapare_variable(file_name: str, params: Dict[str, str]):
    with open(file_name, "r+", encoding="utf-8") as file:
        cfg = file.read()

    def _replacement(matched, params: Dict[str, str]):
        text = str(matched.group(1))
        index = text.find(":")
        default_value = None
        if index > 0:
            var_name = text[0:index]
            default_value = text[index + 1:]
        else:
            var_name = text
        # 获取传入参数
        value = params.get(var_name)
        if value is None and default_value:
            # 设置默认值
            value = default_value

        if (value is None):
            error_msg = "未指定变量'%s'的默认值" % var_name
            raise RuntimeError(error_msg)
        return value

    text = pattern.sub(
        lambda matched: _replacement(matched, params), cfg)
    return text



from __future__ import absolute_import
from .ep.simple_ep_request import SimpleEndPointRequest
from .request_context import WebSocketInvoker, RpcInvoker, PipeRpcInvoker, \
    AbstractInvoker, RpcServiceInfo
from ..rpc import lb as lb_factory

import asyncio


def request_creator(context): return SimpleEndPointRequest(context)


methods = set(['get', 'post', 'put', 'delete', 'patch'])
_request_creator = request_creator


class RpcClient():
    def __init__(self, service: RpcServiceInfo = None):
        self.service = service

    def get_async(self, invoker: RpcInvoker, service: RpcServiceInfo = None):
        '''
        :return Awaitable[Any]
        '''
        return _verb_func('get', self, invoker, service)

    def post_async(self, invoker: RpcInvoker, service: RpcServiceInfo = None):
        '''
        :return Awaitable[Any]
        '''
        return _verb_func('post', self, invoker, service)

    def put_async(self, invoker: RpcInvoker, service: RpcServiceInfo = None):
        '''
        :return Awaitable[Any]
        '''
        return _verb_func('put', self, invoker, service)

    def patch_async(self, invoker: RpcInvoker, service: RpcServiceInfo = None):
        '''
        :return Awaitable[Any]
        '''
        return _verb_func('patch', self, invoker, service)

    def delete_async(self, invoker: RpcInvoker, service: RpcServiceInfo = None):  # noqa
        '''
        :return Awaitable[Any]
        '''
        return _verb_func('delete', self, invoker, service)

    def stream_async(self, invoker: PipeRpcInvoker, service: RpcServiceInfo = None):  # noqa
        '''
        :return Awaitable[httpx.Response]
        '''
        if self.service and service:
            raise RuntimeError('已绑定serivce，不允许再传入新的service')

        if self.service:
            # 当前rpc对象已绑定service
            return rpc.stream_async(invoker, self.service)

        method = invoker.method
        assert method, '使用stream方法，invoker.method不能为空'

        method = method.lower()
        assert methods.__contains__(method), '不支持的方法:' + method
        return _do_request(invoker, service, 'stream')

    def get(self, invoker: RpcInvoker, service: RpcServiceInfo = None):
        return asyncio.run(self.get_async(invoker, service))

    def post(self, invoker: RpcInvoker, service: RpcServiceInfo = None):
        return asyncio.run(self.post_async(invoker, service))

    def put(self, invoker: RpcInvoker, service: RpcServiceInfo = None):
        return asyncio.run(self.put_async(invoker, service))

    def patch(self, invoker: RpcInvoker, service: RpcServiceInfo = None):
        return asyncio.run(self.patch_async(invoker, service))

    def delete(self, invoker: RpcInvoker, service: RpcServiceInfo = None):
        return asyncio.run(self.delete_async(invoker, service))

    def ws(self, invoker: WebSocketInvoker, service: RpcServiceInfo = None):
        '''
        :return websocket.WebSocket
        '''
        if self.service and service:
            raise RuntimeError('已绑定serivce，不允许再传入新的service')

        if self.service:
            # 当前rpc对象已绑定service
            return rpc.ws(invoker, self.service)

        return _do_request(invoker, service, 'ws')

    def stream(self, invoker: PipeRpcInvoker, service: RpcServiceInfo = None):
        '''
        :return httpx.Response
        '''
        return asyncio.run(self.stream_async(invoker, service))

    def bind(self, service: RpcServiceInfo):
        '''
        需要调用的服务的信息
        '''
        return RpcClient(service)

    def set_creator(self, creator):
        '''
        :param creator
        :type 'creator' context => EndPointRequest
        '''
        if not creator:
            raise RuntimeError('creator不能为空值')

        _request_creator = creator  # noqa F841


def _verb_func(
        method: str, target: RpcClient, invoker: RpcInvoker,
        service: RpcServiceInfo):
    '''
    :return Awaitable[Any]
    '''
    if target.service and service:
        raise RuntimeError('已绑定serivce，不允许再传入新的service')
    if target.service:
        invoker.method = method
        # 当前rpc对象已绑定service
        return _do_request(invoker, target.service, 'execute')

    invoker.method = method
    return _do_request(invoker, service, 'execute')


def _do_request(
        invoker: AbstractInvoker, service: RpcServiceInfo, method: str):
    retry = invoker.retry if hasattr(invoker, 'retry') else 0
    lb = lb_factory.create(retry, service)
    context = lb.create_context(invoker, service)
    ep_request = _request_creator(context)

    return lb.request(method, ep_request)


rpc = RpcClient()



from __future__ import absolute_import
from .ep_request import EndPointRequest
from ..request_context import EndPoint
from .internal_rpc import execute as rpc_execute, \
    stream as rpc_stream, ws as rpc_ws


class SimpleEndPointRequest (EndPointRequest):
    async def execute(self, ep: EndPoint):
        resultData = await rpc_execute(self.context, ep)
        return resultData

    async def stream(self, ep: EndPoint):
        '''
        :return Awaitable[httpx.Response]
        '''
        resultStream = await rpc_stream(self.context, ep)
        return resultStream

    def ws(self, ep: EndPoint):
        '''
        :return Stream
        '''
        conn = rpc_ws(self.context, ep)
        return conn



from __future__ import absolute_import
import socket
import netifaces
from log4py import logging
from .local_discovery_client import LocalDiscoveryClient
from .eureka_discovery_client import EurekaDiscoveryClient
from ..config import configure

log = logging.getLogger('python.eureka.discoveryClient')


def getIPAdress():
    '''
    获取本机ip
    '''

    result = []
    for iface in netifaces.interfaces():
        alias = netifaces.ifaddresses(iface)
        ipv4 = alias.get(netifaces.AF_INET)
        if ipv4 is None:
            continue
        for addr_info in ipv4:
            address = addr_info["addr"]
            if address == '127.0.0.1':
                continue
            result.append(addr_info["addr"])

    if len(result) == 0:
        raise RuntimeError('未找到合适的ipv4地址')

    if len(result) > 1:
        log.warning('找到多个符合条件的ipv4地址: %s' % result)

    selected_address = result[0]
    log.info('当前使用的ip地址: ' + selected_address)

    return selected_address


def __create_discovery_client():
    eurekaConfig = configure.get_config('eureka', {'enabled': True})
    DiscoveryClientCls = EurekaDiscoveryClient if eurekaConfig["enabled"] \
        else LocalDiscoveryClient

    ipAddress = getIPAdress()
    hostName = socket.gethostname()
    return DiscoveryClientCls(ipAddress, hostName)


discovery_client = __create_discovery_client()



from __future__ import absolute_import
from log4py import logging
from ...discovery import discovery_client
from .loadbalancer import LoadBalancer
from ..request_context import InternalRequestContext, RpcInvoker, \
    RpcServiceInfo, EndPoint
from ..ep.ep_request import EndPointRequest
from ..rule.simple_rule import SimpleRule

log = logging.getLogger('python.cloud.rpc.loadbalancer')


class DiscoveryServiceLoadBalancer (LoadBalancer):
    '''
    通过Eureka服务动态获取可用的服务端地址的实现
    '''

    def __init__(self, service_id, app_name):
        super().__init__()
        self.service_id = service_id
        self.app_name = app_name
        self._rule = SimpleRule(service_id, discovery_client.client)

    def check_alive(self) -> bool:
        # 重新计算各节点权重值
        try:
            if self._rule.check_alive():
                self._rule.compute()
                return True
        except Exception as err:
            log.error('Error calculating server weights', err)

        # 目标服务例不在线或者计算权重失败时都返回false
        return False

    def chooseOne(self) -> EndPoint:
        inst = self._rule.choose_server()
        return inst

    def create_context(self,
                       invoker: RpcInvoker,
                       service: RpcServiceInfo) -> InternalRequestContext:
        return InternalRequestContext(invoker, service, self.app_name)

    async def execute(self, req: EndPointRequest):
        ep = self._rule.choose_server()
        result_data = await req.execute(ep)
        return result_data

    async def stream(self, req: EndPointRequest):
        ep = self._rule.choose_server()
        result_stream = await req.stream(ep)
        return result_stream

    def ws(self, req: EndPointRequest):
        ep = self._rule.choose_server()
        conn = req.ws(ep)
        return conn



import re
from typing import Dict, Any, List
import functools
from log4py import logging
from .config_value import as_crypto_value, as_password_value

log = logging.getLogger('python.lib.configure.merge')
CRYPTO_CONST = '{crypto}'


ARRAY_INDEX_PATTERN = re.compile("\\[\\d+\\]$")
INTERNAL_PARAM_PATTERN = re.compile("\\$\\(([.\\w-]+)(#(\\w+))??\\)")


def config_merge(target: Dict[str, Any], source: Dict[str, Any], type: str):
    '''
    添加配置项，每一个同名的配置项都对应着一个数组，每一个数组成员的type都不同，来自不同的配置源
    :param key 配置项的key
    :param value 配置项的值
    :param type 配置项的类型，表示来自哪一种数据源
    '''
    for name, src_value in source.items():
        if src_value is None:
            continue

        index = None
        if ARRAY_INDEX_PATTERN.search(name):
            # 形如下面格式的数组
            # a.b.c[0]: 123
            # a.b.c[1]: 456
            start = name.rindex('[')
            index = name[start + 1:-1]
            index = int(index)
            name = name[0:start]

        target_value = target.get(name)

        if isinstance(index, int):
            # 合并数组中的一项
            # 处理形如下面的配置格式
            #    a.b.c[0]: 'value'
            # 此时取到的name = 'c', index = 0, value = 'value'
            # 找到targe中数组配置的值的对象
            if target_value is None:
                target_value = TypedList()
                target_value.item_type = 'array'
                target[name] = target_value

            _merge_plain_array(target_value, src_value, type, index)
        else:
            if isinstance(src_value, Dict):
                # 最终生成的json对象中非叶子节点都是object类型，不用数组表示
                # 叶子节点每个值都是数组，表示从各个配置源得到的值，最终使用的是索引为0的值
                # 非叶子节点
                if target_value is None:
                    target_value: Dict[str, Any] = {}
                    target[name] = target_value
                elif not isinstance(target_value, Dict):
                    error_msg = "配置段类型不匹配. source:%s, target:%s" % (
                        src_value, target_value)
                    raise RuntimeError(error_msg)
                config_merge(target_value, src_value, type)
            else:
                # 叶子节点或json格式的数组对象
                src_type = 'array' if isinstance(src_value, List) else 'simple'
                if target_value is None:
                    target_value = TypedList()
                    target[name] = target_value
                    target_value.item_type = src_type
                try:
                    target_type = target_value.item_type
                    if target_type != src_type:
                        error_msg = "合并的类型不一致, name: %s, sourceType: %s, sourceType: %s" % ( # noqa
                            name, target_type, src_type)
                        raise RuntimeError(error_msg)

                    # 插入到最前面
                    target_value.insert(0, {
                        'type': type,
                        'value': src_value
                    })
                except Exception:
                    log.error(
                        "合并配置时发生错误：targetValue: %s, source=%s, target=%s" % (
                            target_value, source, target)
                    )
                    raise


def _merge_plain_array(target_value, source_value, type: str, index: int):
    '''
    :param targetValue
    :param sourceValue
    :param type 配置所属的类型
    :param index
    '''
    array_value = None
    # 找到type 相同的项
    for config_item in target_value:
        if config_item["type"] == type:
            array_value = config_item.get("value")

    if array_value is None:
        array_value = TypedList()
        target_value.insert(0, {
            'type': type,
            'value': array_value
        })

    if len(array_value) != index:
        raise RuntimeError('合并数组配置必须按下标顺序合并，当前数组长度为%d，要合并的下标为%d' %
                           (len(array_value), index))
    if isinstance(source_value, Dict):
        val: Dict[str, Any] = {}
        # 如果数组中每一项都是对象，递归处理，避免最终结果中出现带[0]这样的属性
        config_merge(val, source_value, type)
        array_value.append(val)
    else:
        array_value.append(source_value)


def process_source(source: Dict[str, Any], params: Dict[str, str]):
    config: Dict[str, Any] = {}
    for key in source:
        c = config
        segments = key.split('.')
        while len(segments) > 1:
            propName = segments.pop(0)
            val = c.get(propName)
            if val is None:
                val: Dict[str, Any] = {}
                c[propName] = val
            c = val

        val = source.get(key)
        c[segments[0]] = val if not isinstance(
            val, str) else _resolve_internal_variable(val, params)
    return config


def _resolve_internal_variable(content: str, params: Dict[str, str]):
    if (content == '' or content is None):
        return content

    def replacement_fn(args):
        expr = args.group(0)
        key = args.group(1)
        param_type = args.group(3)
        if param_type is None:
            param_type = 'string'

        value = params.get(key)
        if value is None:
            raise RuntimeError("未指定变量'%s'的值" % key)

        val_type = 'string'
        if isinstance(value, bool):
            val_type = 'boolean'
        elif isinstance(value, int):
            val_type = 'number'

        if val_type != param_type and \
                (param_type != 'password' or val_type != 'string'):
            raise RuntimeError(
                "%s类型参数的传入值不正确. key: %s, type: %s" % (
                    param_type, key, val_type)
            )
        wrap = value
        if (val_type == 'string'):
            if (value.startswith(CRYPTO_CONST)):
                input = value[CRYPTO_CONST.length:]
                wrap = as_crypto_value(input)
            wrap = as_password_value(
                wrap) if param_type == 'password' else wrap

        replements.append({'expr': expr, 'key': key, 'value': wrap})
        return str(value)

    replements = []
    result_value = INTERNAL_PARAM_PATTERN.sub(replacement_fn, content)

    # 1. content不含参数时，result_value 就是content，类型也是string
    # 2. content包含多于一个参数，只能当string类型处理
    if len(replements) == 1:
        # 等于于一个参数
        item = replements[0]
        if content == item['expr']:
            # 参数与表达式完全匹配，可以直接返回参数对应的类型
            if result_value != item['value']:
                # 一般只会出现在value类型非字符串的时候
                return item['value']

    return result_value


class TypedList(List):
    def __init__(self, *args):
        super().__init__(args)
        self.item_type: str = None


def resolve_params(target: dict, params: dict):
    if params is None:
        return
    param_refs = list()
    keys = list(params.keys())
    for key in keys:
        value = params.get(key)
        if not isinstance(value, str):
            target[key] = value
            continue
        matcher = INTERNAL_PARAM_PATTERN.search(value)
        if matcher is None:
            target[key] = value
            continue

        # 找出参数引用关系
        item = {'key': key, 'refs': list()}
        param_refs.append(item)
        while matcher is not None:
            item['refs'].append(matcher.group(1))
            matcher = INTERNAL_PARAM_PATTERN.search(value, matcher.end())

    def cmp(a, b):
        if b['key'] in a['refs']:
            return 1
        if a['key'] in b['refs']:
            return -1
        return 0

    param_refs.sort(key=functools.cmp_to_key(cmp))

    for param_ref in param_refs:
        key = param_ref['key']
        expr = params.get(key)
        value = _resolve_internal_variable(expr, target)
        target[key] = value



from __future__ import absolute_import

from abc import ABC, abstractmethod, abstractproperty
from ..request_context import AbstractInvoker, RpcServiceInfo, \
    RequestContext, EndPoint
from ..ep.ep_request import EndPointRequest


class LoadBalancer(ABC):
    @abstractmethod
    def create_context(self,
                       invoker: AbstractInvoker,
                       service: RpcServiceInfo) -> RequestContext:
        pass

    @abstractproperty
    def check_alive() -> bool:
        pass

    @abstractmethod
    def chooseOne(self) -> EndPoint:
        return None

    def request(self, method: str, req: EndPointRequest):
        if method == 'execute':
            return self.execute(req)
        elif method == 'stream':
            return self.stream(req)
        elif method == 'ws':
            return self.ws(req)  # 底层创建websocket用的是同步方法，当前返回值会转换成异步调用的包装
        raise RuntimeError('不支持的方法:' + method)

    @abstractmethod
    async def execute(self, req: EndPointRequest):
        pass

    @abstractmethod
    async def stream(self, req: EndPointRequest):
        pass

    @abstractmethod
    def ws(self, req: EndPointRequest):
        '''
        :return websocket.WebSocket
        '''
        pass



from __future__ import absolute_import
from .loadbalancer import LoadBalancer
from ..request_context import RpcInvoker, RpcServiceInfo, \
    EndPoint, RequestContext
from ..ep.ep_request import EndPointRequest


class RetryLoadBalancer(LoadBalancer):

    def __init__(self, lb: LoadBalancer):
        super().__init__()
        self._lb = lb

    def check_alive(self):
        return self._lb.check_alive()

    def chooseOne(self) -> EndPoint:
        return self._lb.chooseOne()

    def create_context(self,
                       invoker: RpcInvoker,
                       service: RpcServiceInfo) -> RequestContext:
        return self._lb.create_context(invoker, service)

    async def execute(self, req: EndPointRequest):
        result_data = await self._lb.execute(req)
        return result_data

    async def stream(self, req: EndPointRequest):
        result_stream = await self._lb.stream(req)
        return result_stream

    def ws(self, req: EndPointRequest):
        conn = self._lb.ws(req)
        return conn



from typing import Dict, List
import json


class WebError(RuntimeError):
    def __init__(self,
                 message: str,
                 code: str = None,
                 desc: str = None,
                 headers: Dict[str, str] = {},
                 status: int = 400):
        super().__init__(message)
        self.code = code
        self.desc = desc
        self.headers = headers
        self.status = status
        self.details: List[str] = []

    def __dir__(self):
        # 返回自定义的属性和方法列表
        return super.__dir__() + \
            ['code', 'desc', 'headers', 'status', 'details']

    def __str__(self):
        obj = dict(self.__dict__)
        obj['message'] = super().__str__()
        return json.dumps(obj, ensure_ascii=False)



from __future__ import absolute_import
from log4py import Logger
import os
import sys
from typing import Dict
from ._configure import Configure


def _load_env_params(params: Dict[str, str]):
    '''
    初始化环境变量的配置
    '''
    for envKey, value in os.environ.items():
        if envKey.find('NODE_') == 0:
            key = envKey[envKey.find('_') + 1:].lower().replace('_', '.')
            params[key] = value


def _load_cmd_params(params: Dict[str, str]):
    '''
    初始化命令行的配置参数
    '''
    for arg in sys.argv:
        if arg.find('--') == 0:
            info = arg[2:].split('=')
            if len(info) == 2:
                [key, value] = info
                params[key.lower()] = value


_params: Dict[str, str] = {}
# 获取环境变量
_load_env_params(_params)
# 获取命令行参数，可以保证命令行参数优先于环境变量
_load_cmd_params(_params)

Logger.set_level('INFO')

configure = Configure(_params)



import json
import os
import base64
import pyDes

work_dir = os.getcwd()
# 加载密钥信息
crypto_file = os.path.join(work_dir, 'crypto.json')

if os.path.exists(crypto_file):
    with open(crypto_file, "r+") as fn:
        cryptoParams = json.load(fn)

if not os.path.exists(crypto_file):
    def normalize_value(input: str):
        if input.startswith(CryptoObject.CRYPTO_CONST):
            raise RuntimeError('未找到配置文件解密密钥')
        return input
else:
    def normalize_value(input: str):
        if not input.startswith(CryptoObject.CRYPTO_CONST):
            return input
        return CryptoObject(input, lambda x: cryptoParams)

CRYPTO_CONST = '{crypto}'


def as_password_value(input: str):
    return PasswordObject(input)


def as_crypto_value(input: str):
    if cryptoParams is None:
        raise RuntimeError('未找到配置文件解密密钥')
    return CryptoObject(input, lambda x: cryptoParams)


class SecurityObject(object):
    def __init__(self, data: str):
        super().__init__()
        self._data = data

    def toString(self):
        raise RuntimeError('not implemented')

    def toJSON(self):
        return '***************'


class CryptoObject (SecurityObject):
    def __init__(self, data: str, cb):
        super().__init__(data)
        self._cb = cb

    def __str__(self):
        o = self._cb()
        hexKey = base64.decodestring(o['key'])
        hexIV = base64.decodestring(o['iv'])
        cipher_data = base64.decodestring(self._data)
        algo = pyDes.triple_des(
            hexKey, pyDes.CBC, IV=hexIV, pad=None, padmode=pyDes.PAD_PKCS5)
        plain_text = str(algo.decrypt(cipher_data), 'utf-8')
        return plain_text

    # __repr__ = __str__


class PasswordObject(SecurityObject):
    def __init__(self, data: str):
        super().__init__(data)

    def __str__(self):
        return str(self._data)

    # __repr__ = __str__



from __future__ import absolute_import
import httpx
from functools import partial

from ..web_error import WebError
from ..request_context import RequestContext, EndPoint
from ...discovery import discovery_client as client


def _create_request_option(context: RequestContext, url: str):
    option = {
        'method': context.invoker.method,
        'url': url,
        'data': context.invoker.body,
        'body': context.invoker.body,
        'headers': {
            'accept': 'application/json,*/*',
            'accept-encoding': 'gzip, deflate',
            'user-agent': 'nodejs rest client',
            'content-type': 'application/json',
            # 表示异步调用
            'x-client-ajax': 'true'
        }.update(context.headers),
        'timeout': 1 if not context.timeout else (context.timeout / 1000)
    }
    context.process_request_option(option)
    return option


async def _walk_using_urllib(ctx, url):
    option = _create_request_option(ctx, url)
    async with httpx.AsyncClient() as client:
        res: httpx.Response = await client.request(
            method=option["method"],
            url=option["url"],
            data=option["body"],
            headers=option["headers"],
            timeout=option["timeout"]
        )
        return res


async def _do_request(context, ep) -> httpx.Response:
    if ep:
        url = "http://%s:%d%s" % (ep.host, ep.port, context.path_and_query)
        res = await _walk_using_urllib(url)
    else:
        res = await client.client.walk_nodes(
            app_name=context.service.service_id,
            service=context.path_and_query,
            walker=partial(_walk_using_urllib, context)
        )

    if res.status_code >= 400:
        timeout = context.timeout if context.timeout else -1
        service = context.service
        invoker = context.invoker

        message = "%s --> %s --> [timeout:%ds, cost: %fs] [%s] [%s] %s" % (
            res.reason_phrase, res.text, timeout, res.elapsed.total_seconds(),  # noqa
            service.service_id, invoker.method, context.url)
        err = WebError(message, status=res.status_code)
        context.resolve_error(res, err)
        raise err

    return res


async def execute(context: RequestContext, ep: EndPoint):
    '''
    :return Awaitable[Any]
    '''

    res: httpx.Response = await _do_request(context, ep)
    content_type = res.headers['content-type'] if res.headers['content-type'] \
        else ''

    if context.invoker.json is not False and \
            type(content_type) == str and content_type.find('json') >= 0:
        return res.json()
    return res.text


async def stream(context: RequestContext, ep: EndPoint):
    '''
    :return Awaitable[httpx.Response]
    '''

    res: httpx.Response = await _do_request(context, ep)
    return res


def ws(context: RequestContext, ep: EndPoint):
    if ep:
        url = "ws://%s:%d%s" % (ep.host, ep.port, context.path_and_query)
    else:
        ec = client.client
        app_name = context.service.service_id.upper()
        service = context.path_and_query
        node = ec.__get_available_service(app_name)
        url = ec.__generate_service_url(node, False, False)
        if service.startswith("/"):
            url = url + service[1:]
        else:
            url = url + service

    option = _create_request_option(context, url, False)
    from websocket import create_connection
    conn = create_connection(url=option["url"],
                             timeout=option["timeout"],
                             headers=option['headers']
                             )

    return conn



from __future__ import absolute_import
from .loadbalancer import LoadBalancer
from ..request_context import InternalRequestContext, \
    RpcInvoker, RpcServiceInfo, EndPoint
from ..ep.ep_request import EndPointRequest


class DirectLoadBalancer (LoadBalancer):

    '''
    显示指定了调用的目标地址的LoadBalancer实现方式
    '''

    def __init__(self):
        super().__init__()

    def check_alive(self):
        return True

    def chooseOne(self) -> EndPoint:
        return None

    def create_context(self, invoker: RpcInvoker, service: RpcServiceInfo) \
            -> InternalRequestContext:
        return InternalRequestContext(invoker, service, None)

    @property
    def type(self):
        return 'direct'

    async def execute(self, req: EndPointRequest):
        ep = self._getEndPoint(req)
        result_data = await req.execute(ep)
        return result_data

    async def stream(self, req: EndPointRequest):
        ep = self._getEndPoint(req)
        result_stream = await req.stream(ep)
        return result_stream

    def ws(self, req: EndPointRequest):
        ep = self._getEndPoint(req)
        conn = req.ws(ep)
        return conn

    def _getEndPoint(self, req: EndPointRequest):
        target_url = req.context.url
        port = req.context.url_port
        if not port:
            port = 80 if target_url.protocol == 'http:' else 443
        return EndPoint(
            id="%s:%s" % (self.type, req.context.service.url),
            host=req.context.url_host,
            port=port
        )



from __future__ import absolute_import

from collections.abc import Callable
from typing import Dict, Any, Union, Tuple, TypeAlias, Mapping
import re
import io
import httpx
from abc import ABC, abstractmethod
from .web_error import WebError
from urllib.parse import urlparse, urljoin, urlencode, parse_qsl, urlunparse


PARAM_PARTERN = re.compile(':(\\w+)')
HEADER_PARTTERN = re.compile('[A-Z]')


class AbstractInvoker(ABC):
    def __init__(self, path: str,
                 params: Dict[str, Any] = None,
                 query: Dict[str, Any] = None,
                 headers: Dict[str, Any] = None,
                 context: Dict[str, Any] = None
                 ):
        self.path = path
        self.params = params
        self.query = query
        self.headers = headers
        self.context = context


class WebSocketInvoker(AbstractInvoker):
    pass


JsonType: TypeAlias = Any
RpcFallback: TypeAlias = Callable[[BaseException, AbstractInvoker], Any]


class RpcInvoker(AbstractInvoker):
    def __init__(self, path: str,
                 params: Mapping[str, Any] = None,
                 query: Mapping[str, Any] = None,
                 body: Union[JsonType, io.IOBase] = None,
                 headers: Mapping[str, Any] = None,
                 retry: int = None,
                 context: Mapping[str, Any] = None,
                 timeout: int = None,
                 **kwargs
                 ):
        super().__init__(path, params=params, query=query,
                         headers=headers, context=context)
        self.body = body
        # 仅在execute方法内有效，调用返回的类型是否是json,如果为false则返回原始字符串文本
        self.json = kwargs.get('json', True)
        assert isinstance(self.json, bool), "json属性类型应变为bool类型"
        self.retry = retry
        self.timeout = timeout


class PipeRpcInvoker(RpcInvoker):

    def __init__(self, path: str,
                 method: str,
                 params: Mapping[str, Any] = None,
                 query: Mapping[str, Any] = None,
                 body: Union[JsonType, io.IOBase] = None,
                 headers: Mapping[str, Any] = None,
                 retry: int = None,
                 context: Mapping[str, Any] = None,
                 timeout: int = None,
                 **kwargs):
        super().__init__(path, params=params, query=query, body=body,
                         headers=headers, retry=retry,
                         context=context, timeout=timeout, **kwargs)
        self.method = method


class RpcServiceInfo():
    def __init__(self,
                 service_id: str = None,
                 url: str = None,
                 rpc_type: str = None,
                 path: str = None,
                 timeout: int = None
                 ):
        self.service_id = service_id
        self.url = url
        self.path = path
        self.rpc_type = rpc_type
        self.timeout = timeout


class EndPoint():
    def __init__(self, id: str, host: str, port: int):
        self.id = id
        self.host = host
        self.port = port


class RequestContext(ABC):
    def __init__(self, invoker: RpcInvoker, service: RpcServiceInfo,
                 app_name: str):
        self.invoker = invoker
        self.service = service
        self.app_name = app_name

        path = invoker.path
        query = invoker.query if invoker.query is not None else {}
        params = invoker.params if invoker.params is not None else {}

        headers = dict(invoker.headers) if invoker.headers else {}
        if invoker.context:
            self._add_headers(headers, invoker.context)
        self.headers = headers

        base_url, path_and_query = self._get_path_and_query(path)
        parse_result = urlparse(urljoin(base_url, path_and_query))
        query_list = parse_qsl(parse_result.query)
        for name, value in query.items():
            query_list.append((name, value))

        def _replacement(matched, params):
            var_name: str = matched.group(1)
            # 获取传入参数
            value = params.get(var_name)
            if (value is None):
                error_msg = "不存在的属性：'%s'" % var_name
                raise RuntimeError(error_msg)
            return str(value)

        new_path = PARAM_PARTERN.sub(
            lambda matched: _replacement(matched, params), parse_result.path)
        new_qs = urlencode(query_list)
        self.url = urlunparse([
            parse_result.scheme,
            parse_result.netloc,
            new_path,
            "",
            new_qs, ""])
        self.url_port = parse_result.port
        self.url_host = parse_result.hostname
        self.path_and_query = new_path + "?" + new_qs

    @property
    def timeout(self):
        if self.invoker.timeout:
            return self.invoker.timeout
        else:
            return self.service.timeout

    def process_request_option(self, option: Dict[str, Any]):
        # 什么也不做
        pass

    @abstractmethod
    def resolve_error(self, err: BaseException):
        pass

    @abstractmethod
    def _get_path_and_query(self, path: str) -> Tuple[str, str]:
        '''
        :return [baseUrl: str, path_and_query: str]
        '''
        pass

    @abstractmethod
    def _add_headers(self, headers: Dict[str, Any], context: Dict[str, Any]):
        pass

    def _to_header_name(self, name: str, prefix: str):
        key = HEADER_PARTTERN.sub(
            lambda matched: '-' + matched.group(0).upper(), name)

        # 补充'x-'前缀
        if key.startswith('-'):
            return prefix + key
        return prefix + '-' + key


class InternalRequestContext(RequestContext):
    def _get_path_and_query(self, path: str):
        base_url = self.service.url
        if not base_url:
            base_url = "http://%s" % self.service.service_id

        path_prefix = self.service.path if self.service.path else '/'
        path_and_query = urljoin(path_prefix, path)
        return (base_url, path_and_query)

    def _add_headers(self, headers: Dict[str, Any], context: Dict[str, Any]):
        # const principal = {}
        # let extras

        # const webContext = $asynctx.webContext
        # if (webContext) {
        #     // 当前调用是在处理web请求的时候发生的
        #     if (webContext.principal) {
        #         // website身份认证信息
        #         principal.id = webContext.principal.id
        #         if (!principal.userId & & webContext.principal.id) {
        #             principal.userId = webContext.principal.id
        #         }
        #         principal.tenantId = webContext.principal.tenantId
        #         principal.tenantCode = webContext.principal.tenantCode

        #         // TODO: 应该删除掉
        #         principal.orgId = webContext.principal.orgId
        #         principal.applicationId = webContext.principal.applicationId
        #     }
        #     extras = webContext.extras
        # }

        # if (!principal.tenantId) {
        #     // 当前调用是在消息队列或定时任务等其它非web请求中调用的
        #     // 消息队列处理器中中没有webContext对象
        #     principal.tenantId = $asynctx.tenantId
        # }

        for name, value in context:
            if value is None:
                continue
            key = self._to_header_name(name, 'x')
            headers[key] = context[name]

        if self.app_name:
            headers['i-rpc-client'] = self.app_name

        # TODO: 用于调用链跟踪的信息
        # for name, value in asynctx.tracing:
        #     headers[name] = asynctx.tracing

    def resolve_error(self, res: httpx.Response, err: WebError):
        try:
            data = res.json()
        except Exception:
            # 什么也不做
            data = {
                'code': 'json_format_error',
                'desc': res.text
            }
        # 默认json格式
        err.code = data.get("code")
        err.desc = data.get("desc")
        err.details = data.get("details")



from __future__ import absolute_import
from typing import Dict
from urllib.parse import urlparse
from threading import Timer
from ...discovery import discovery_client
from ...config import configure
from ..request_context import RpcServiceInfo, EndPoint
from .direct_loadbalancer import DirectLoadBalancer
from .discovery_service_loadbalancer import DiscoveryServiceLoadBalancer
from .local_service_loadabalancer import LocalServiceLoadBalancer
from .loadbalancer import LoadBalancer
from .retry_loadbalancer import RetryLoadBalancer
from mctech_actuator.health import manager

# gateway = GatewayLoadBalancer()
direct = DirectLoadBalancer()


load_balance_cache: Dict[str, LoadBalancer] = {}
local_service: Dict[str, str] = {}
dependencies: Dict[str, bool] = {}


def _metric_endpoint(*, request):
    return [v for v in load_balance_cache.keys()]


manager.add_metric({
    # 相对于/actuator/metrics的路径，必须以'/'开头
    'path': '/dependencies',
    'endpoint': _metric_endpoint
})

CHECK_ALIVE_INTERVAL = 5

# 定时检查目标服务的信息是否还有效，如果不存在了，则会移除


def print_time():
    invalidates = []
    for serviceId, lb in load_balance_cache.items():
        if not lb.check_alive():
            invalidates.append(serviceId)

    # 存在失效的服务，从缓存中移除
    if len(invalidates) > 0:
        for service_id in invalidates:
            del load_balance_cache[service_id]


# 指定1秒后执行print_time函数
t = Timer(CHECK_ALIVE_INTERVAL, print_time)
t.start()


app_name: str = None


def _load_config():
    config = configure.get_config('mctech.rpc.service')

    for service_id, service_url in config:
        result = urlparse(service_url)
        local_service[service_id] = EndPoint(
            id=service_id,
            host=result['hostname'],
            port=result.get('port')
        )


def create(retry: int, service: RpcServiceInfo) -> LoadBalancer:

    if not app_name:
        _load_config()

    retry = retry or 0
    rpc_type = service.rpc_type if service.rpc_type else 'internal'

    # lb: LoadBalancer
    # if type == 'gateway':
    #   lb = gateway
    # el
    if rpc_type == 'internal':
        # 配置的是servicieId
        lb = get_internal_loadbalancer(service.service_id)
    else:
        raise RuntimeError('不支持的调用方式')

    if retry > 0:
        lb = RetryLoadBalancer(lb)
    return lb


def get_internal_loadbalancer(service_id: str) -> LoadBalancer:
    if not service_id:
        lb = direct
    else:
        lb = load_balance_cache.get(service_id)
        if not lb:
            inst = local_service.get(service_id)
            if inst:
                lb = LocalServiceLoadBalancer(inst, app_name)
            elif not discovery_client.isLocal:
                lb = DiscoveryServiceLoadBalancer(service_id, app_name)
            elif not inst:
                raise RuntimeError(
                    "本地调用配置中未找到service_id为: '%s'的服务" % service_id
                )

            load_balance_cache[service_id] = lb
            dependencies[service_id] = True
    return lb



from __future__ import absolute_import
from ..config import configure


class LocalDiscoveryClient():
    def __init__(self, localAddress: str, hostName: str):
        info = configure.get_app_info()
        self.local_instance = {
            'hostName': hostName,
            'app': info.get("name"),
            'ipAddr': localAddress,
            'port': {
                '$': info.get("port")
            }
        }

    @property
    def isLocal(self):
        return True

    @property
    def client():
        raise RuntimeError('LocalDiscoveryClient不支持client属性')

    def start(self):
        # 什么也不做
        pass

    def register(self):
        # 什么也不做
        pass

    def unregister(self):
        # 什么也不做
        pass

    def load_config(self):
        return configure



from __future__ import absolute_import

from .abstarct_rule import AbstractRule


class SimpleRule(AbstractRule):
    def check_alive(self):
        apps = self._eureka.applications.get_application(self._app_id_upper)
        return len(apps.up_instances) > 0

    def _choose(self):
        if not self.check_alive():
            return None

        apps = self._eureka.applications.get_application(self._app_id_upper)
        inst = apps.up_instances[0]
        return inst



from __future__ import absolute_import
import os
import sys
from typing import Dict, Any
from .config_merger import config_merge, process_source, \
    TypedList, resolve_params
from .config_value import SecurityObject
from . import yaml_loader as loader
import json

# 参数配置及获取结果规则表
config_source_info: Dict[str, Any] = {}


class ComplexEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, SecurityObject):
            return obj.toJSON()

        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def _get_config_base_dir(params: Dict[str, str]):
    dir_name = os.path.dirname(sys.argv[0])
    if not os.path.exists(os.path.join(dir_name, 'bootstrap.yml')):
        dir_name = params.pop('config.dir',  sys.path[0])
    return dir_name


def _deep_copy_property(target: Dict[str, Any], source: Dict[str, Any]):
    for name, value in source.items():
        # 只有数组或叶子节点的简单值才会对应着来自不同profile里的值
        # 非叶子结点只是一个连接叶子节点的桥梁，不会产生多个值
        if isinstance(value, list):
            union: TypedList = value
            if union.item_type != "array":
                # 非数组，取第一项的值，直接覆盖
                value = union[0].get("value")
            else:
                # 数组值
                v = []
                for i in range(0, len(union)):
                    arr = union[i].get("value")
                    v.extend(arr)
                value = v
        else:
            if isinstance(value, Dict):
                new_value: Dict[str, Any] = {}
                target[name] = new_value
                _deep_copy_property(new_value, value)
                value = new_value

        target[name] = value


def _create_actual_config():
    # 生成一个合并后的配置
    # 生成代理对象
    config: Dict[str, Any] = {}
    _deep_copy_property(config, config_source_info)
    return config


class Configure():
    def __init__(self, params: Dict[str, str]):
        self.params = params
        # 内置在yaml文件内部的参数 $(param这样的格式)
        self.internal_params: Dict[str, Any] = {}
        self._profile_map: Dict[str, Any] = {}
        self._active_profiles = ""

        # 加载默认的配置
        self.use(
            {
                "application": {"port": 8080, "profiles": {"active": ""}},
                "management": {"contextPath": "/actuator"}
            },
            "default-internal",
            "default"
        )

        # 加载bootstrap.yml文件
        self._config_base_dir = _get_config_base_dir(params)
        self._load_config_from_local(params)
        # 合并"default"配置的默认值
        cfg_list = self._profile_map.get("default")
        if cfg_list and len(cfg_list):
            for config in cfg_list:
                config_merge(target=config_source_info,
                             source=config.get("source"),
                             type=config.get("name"))
                config["merged"] = True

                # 从"default"配置上提取激活的profiles
                active_profiles = _get_config_segment(
                    config.get("source"),
                    "application.profiles.active"
                )
                if active_profiles:
                    if self._active_profiles:
                        self._active_profiles = self._active_profiles + \
                            "," + active_profiles
                    else:
                        self._active_profiles = active_profiles
        self._config = _create_actual_config()

    def use(self, source, name: str, profile: str,
            params: Dict[str, str] = {}):
        '''
        加载从其它来源获取到的配置

        :param source
        :param name: 用于追跟踪配置段来源的标识
        :param profile:
        :param params: 内部引用的参数集合
        '''
        source_list = self._profile_map.get(profile)
        if source_list is None:
            source_list = []
            self._profile_map[profile] = source_list

        # 合并参数
        resolve_params(self.internal_params, params)
        # 替换处理内置参数
        source_json = process_source(source, self.internal_params)

        source_list.append({'name': name, 'source': source_json})

    def merge(self):
        if (self._active_profiles):
            active_profile_names = self._active_profiles.split(",")
            # 加载远程服配置中的"defaualt"配置
            active_profile_names.insert(0, "default")
            # 根据激活的profiles加载其它的配置
            for name in active_profile_names:
                cfg_list = self._profile_map.get(name)
                if (cfg_list is None or len(cfg_list) == 0):
                    continue

                for config in cfg_list:
                    if config.get('merged'):
                        # 之前已经合并过，不再合并
                        continue
                    config_merge(config_source_info,
                                 config["source"], config["name"])
                    config['merged'] = True

        self._config = _create_actual_config()

    @property
    def active_profiles(self):
        return self._active_profiles

    def get_config_source_info(self):
        return config_source_info

    def get_app_info(self):
        app = self._config.get("application")
        return {
            "name": app.get("name"),
            "port": app.get("port"),
            "env": app.get("env"),
            "cloudName": app.get("cloudName")
        }

    def get_config(self, prefix="", config: Dict[str, Any] = None):
        '''
        获取指定配置段的配置值
        :param prefix 配置对象中路径前缀
        :param config 配置的默认值
        '''
        if config is None:
            config = {}

        if (len(prefix) == 0):
            v = self._merge(config, self._config)
            return v

        c = _get_config_segment(self._config, prefix)
        return self._merge(config, c)

    def _merge(self, target: Dict[str, Any], source: Dict[str, Any]):
        if (source is None):
            return target

        for name, source_value in source.items():
            if (source_value is None):
                continue

            # json对象
            if isinstance(source_value, Dict):
                target_value = target.get(name)
                if (target_value is None):
                    target_value: Dict[str, Any] = {}
                    target[name] = target_value
                elif not isinstance(target_value, Dict):
                    error_msg = "配置段类型不匹配. source: %s, target: %s" % (
                        source_value, target_value)
                    raise RuntimeError(error_msg)
                self._merge(target_value, source_value)
            else:
                # 其它值类型
                target[name] = source_value
        return target

    # / **
    #  * @ param {Map < string, string >} params
    #  * /
    def _load_config_from_local(self, params):
        config_names = ["bootstrap", "individual"]
        for name in config_names:
            config_file = os.path.join(self._config_base_dir, name + ".yml")
            abs_file = os.path.abspath(config_file)
            if (not os.path.exists(abs_file)):
                continue
            # 把yaml文件解析为一个或多个配置对象
            profile_configs = loader.get_yaml_objects(config_file, params)
            # 把各个配置对象按profiles -> config的映射方式存到map中
            for cfg in profile_configs:
                profile_name = cfg.pop("application.profiles", "default")
                params: Dict[str, str] = cfg.pop("params", {})
                self.use(cfg, profile_name + '-local', profile_name, params)


def _get_config_segment(config: Dict[str, Any], prefix: str):
    c = config
    segments = prefix.split('.')
    for i in range(0, len(segments)):
        segment = segments[i]
        c = c.get(segment)
        if (c is None):
            break
    return c



from __future__ import absolute_import
from abc import ABC, abstractmethod
from ..request_context import RequestContext, EndPoint


class EndPointRequest(ABC):
    def __init__(self, context: RequestContext):
        self.context = context

    @abstractmethod
    async def execute(self, ep: EndPoint):
        '''
        :return Awaitable[Any]
        '''
        pass

    @abstractmethod
    async def stream(self, ep: EndPoint):
        '''
        :return Awaitable[httpx.Response]
        '''
        pass

    @abstractmethod
    def ws(self, ep: EndPoint):
        '''
        :return websocket.WebSocket
        '''
        pass



from __future__ import absolute_import
from .loadbalancer import LoadBalancer
from ..request_context import InternalRequestContext, RpcInvoker, \
    RpcServiceInfo, EndPoint
from ..ep.ep_request import EndPointRequest


class LocalServiceLoadBalancer (LoadBalancer):
    '''
    显示指定了调用的目标地址的LoadBalancer实现方式
    '''

    def __init__(self, inst: EndPoint, app_name):
        super().__init__()
        self._inst = inst
        self.app_name = app_name

    def check_alive(self):
        return True

    def chooseOne(self) -> EndPoint:
        return self._inst

    def create_context(self,
                       invoker: RpcInvoker,
                       service: RpcServiceInfo) -> InternalRequestContext:
        return InternalRequestContext(invoker, service, self.app_name)

    async def execute(self, req: EndPointRequest):
        result_data = await req.execute(self._inst)
        return result_data

    async def stream(self, req: EndPointRequest):
        result_stream = await req.stream(self._inst)
        return result_stream

    def ws(self, req: EndPointRequest):
        conn = req.ws(self._inst)
        return conn
